/*set CALCTYPE to 0 to compile for the 89, 1 to compile for the 92+, and 2 to compile for the v200. I use
this way because the program options screen is buried too deeply for my tastes*/
#define CALCTYPE 2
#if CALCTYPE == 0
#define USE_TI89
#elif CALCTYPE == 1
#define USE_TI92PLUS
#else
#define USE_V200
#endif

#define OPTIMIZE_CALC_CONSTS

#define RELEASE 2 /*The revision that this program is on*/
#define RANDOM_STRENGTH 4
#define DATA_NAME SYMSTR("sweepdat")
#define QUEUE_SIZE 128

#define calc(x, y) PSEUDO_CONST_CALC(y, x, y, y)

#define GRAY_MEM_HW1 3848
#define GRAY_MEM_HW2 7688
#define MAX_HEIGHT calc(16, 20)
#define MAX_WIDTH calc(26, 39)

#include <tigcclib.h>
#include "extgraph.h"
#include "extrand.h"
#include "vtilog.h"

typedef enum
{
	CHECK_MISS,
	CHECK_HIT
} CHECK;

typedef enum
{
	DIFFICULTY_BEGINNER,
	DIFFICULTY_INTERMEDIATE,
	DIFFICULTY_ADVANCED,
	DIFFICULTY_CUSTOM,
	DIFFICULTY_PROGRESSIVE
} DIFFICULTY;

typedef enum
{
	GAME_WIN,
	GAME_LOSE,
	GAME_SAVE,
	GAME_NO_GRAY
} GAME;

typedef enum
{
	HIGHSCORES_TIMED,
	HIGHSCORES_SCORED
} HIGHSCORES;

typedef enum
{
	INPUT_ESC,
	INPUT_SAVE,
	INPUT_CHECK,
	INPUT_FLAG,
	INPUT_CHECKAROUND
} INPUT;

typedef enum
{
	MAIN_MENU_ESC,
 	MAIN_MENU_START
} MAIN_MENU_RETURN;

typedef enum
{
	MOVEMENT_FLYING,
	MOVEMENT_WALKING
} MOVEMENT;

typedef enum
{
	NUMBERING_TORUS,
	NUMBERING_PLANE
} NUMBERING;

typedef enum
{
	SCORING_NONE,
	SCORING_TIMED,
	SCORING_SCORED
} SCORING;

typedef enum
{
	SPRITE_0,
	SPRITE_1,
	SPRITE_2,
	SPRITE_3,
	SPRITE_4,
	SPRITE_5,
	SPRITE_6,
	SPRITE_7,
	SPRITE_8,
	SPRITE_CURSOR,
	SPRITE_SQUARE,
	SPRITE_MINE,
	SPRITE_MINE_HIT,
	SPRITE_FLAG
} SPRITE;

/*structure to store information about each square*/
typedef struct
{
	unsigned char adjacent;
	unsigned char checked;
	unsigned char flag;
	unsigned char mine;
} FLAGS;

typedef struct
{
	unsigned char column;
	unsigned char delay;
	unsigned char height;
	unsigned char hw;
	/*the row where you select high scores, custom setup, or key timing*/
	unsigned char misc;
	unsigned char row;
	unsigned char width;
	unsigned short mines;
	DIFFICULTY difficulty;
	MOVEMENT movement;
	NUMBERING numbering;
	SCORING scoring;
} GAME_SETTINGS;

typedef struct
{
	short x;
	short y;
} LOCATION;

typedef struct
{
	short front;
	short back;
	LOCATION data[QUEUE_SIZE];
} QUEUE;

typedef struct
{
	unsigned char flags;
	unsigned short emptySquaresCleared;
	unsigned short level;
	unsigned short nonZeroSquares;
	unsigned short squaresCleared;
	unsigned long score;
	LOCATION cursor;
} GAME_STATE;

typedef struct
{
	unsigned char min;
	unsigned char sec;
	unsigned char ms;
} TIME;

typedef struct
{
	/*need light plane only*/
	unsigned long backgroundSave[5];
	TIME time;
	BOOL print;
	LOCATION coords;
} TIMER;

typedef struct
{
	unsigned char row;
	unsigned long scores[2];
	MOVEMENT movement;
	NUMBERING numbering;
	SCORING scored;
	TIME times[3][3];
} HIGH_SCORES;

typedef struct
{
	unsigned char calc;
	unsigned char hw;
	unsigned char initials[3];
	unsigned char random_strength;
} FILE_SIG;

typedef struct
{
	struct
	{
		unsigned short size;
		FILE_SIG sig;
	} header;
	struct
	{
		unsigned char mines[3];
		unsigned long seed;
		BOOL savedGame;
		GAME_SETTINGS settings;
		GAME_STATE state;
		HIGH_SCORES highScores;
		RANDOM_BUFFER randomState;
		TIME time;
	} data;
	struct
	{
		FLAGS minefield[MAX_HEIGHT][MAX_WIDTH];
	} secretData;
	struct
	{
		unsigned char hash[16];
		unsigned char padding;
		unsigned char type[7];
	} footer;
} DATA_FILE;

typedef struct
{
	char backgroundSave[2][18][7];
	LCD_BUFFER bufferMain;
	LCD_BUFFER bufferMainMenu;
	QUEUE queue;
} MEM_BLOCK;

static DATA_FILE *LoadDataFile(HANDLE *fileHandle);
static void CloseDataFile(DATA_FILE *, HANDLE fileHandle);
static MAIN_MENU_RETURN MainMenu(unsigned char[], GAME_SETTINGS *, HIGH_SCORES *, LCD_BUFFER);
static void MainMenuText(unsigned char[], GAME_SETTINGS);
static void MainMenuHighlights(unsigned char, unsigned char[], unsigned char[], GAME_SETTINGS);
static void MainMenuDrawProgressive(char, char[], GAME_SETTINGS);
static void MainMenuDrawDot(GAME_SETTINGS);
static void MainMenuDrawHighlight(unsigned char[], GAME_SETTINGS);
static void HighScoreMenu(HIGH_SCORES *);
static void HighScoreText(HIGH_SCORES);
static void HighScoreHighlights(HIGH_SCORES);
static void HighScoreDrawHighlight(HIGH_SCORES);
static void HighScoreDrawDot(HIGH_SCORES);
static void HighScoreDrawScores(HIGH_SCORES);
static void CustomMenu(GAME_SETTINGS *);
static void CustomMenuText(GAME_SETTINGS);
static void CustomMenuHighlight(GAME_SETTINGS);
static void KeyDelayMenu(GAME_SETTINGS *);
static void KeyDelayMenuText(GAME_SETTINGS);
static void KeyDelayMenuTestStrip(void);
static void KeyDelayMenuHighlight(unsigned char);
static void SetupGameParameters(unsigned char [], GAME_SETTINGS *, GAME_STATE *);
static void SetupMinefield(FLAGS[][MAX_WIDTH], GAME_SETTINGS, GAME_STATE *);
static void Wrap(LOCATION *, GAME_SETTINGS);
static GAME PlayGame(BOOL *, FLAGS[][MAX_WIDTH], GAME_SETTINGS, GAME_STATE *, char[][18][7], QUEUE *,
					 unsigned char[]);
static void DrawMinefield(GAME_SETTINGS, FLAGS[][MAX_WIDTH]);
static void DrawBorder(GAME_SETTINGS);
static void DrawSprite(SPRITE, GAME_SETTINGS, LOCATION);
static void DrawInfoBox(GAME_SETTINGS, GAME_STATE, char[][18][7]);
static SCR_RECT LocateInfoBox(GAME_SETTINGS, GAME_STATE);
static void EraseInfoBox(GAME_SETTINGS, GAME_STATE, char[][18][7]);
static INPUT GetInput(GAME_SETTINGS, GAME_STATE *, char[][18][7]);
static CHECK CheckSquare(FLAGS[][MAX_WIDTH], GAME_SETTINGS, GAME_STATE *, QUEUE *);
static void DrawMines(FLAGS[][MAX_WIDTH], GAME_SETTINGS, GAME_STATE);
static void DoFlag(FLAGS[][MAX_WIDTH], GAME_SETTINGS, GAME_STATE *);
static CHECK DoCheckAround(FLAGS[][MAX_WIDTH], GAME_SETTINGS, GAME_STATE *, QUEUE *);
static CHECK DoAutoClear(FLAGS[][MAX_WIDTH], GAME_SETTINGS, GAME_STATE *, QUEUE *);
static BOOL CheckForWin(GAME_SETTINGS, GAME_STATE);
static void PrintTime(void);
static void CheckHighScores(GAME_SETTINGS, HIGH_SCORES *, unsigned long);
void SleepMS_HW1(unsigned short time);
void SleepMS_HW2(unsigned short time);

volatile TIMER Timer;

DEFINE_INT_HANDLER(Int5Timer)
{
	/*Each one of these odd casts is to get rid of the volatile modifier for the scope of the interrupt
	handler, which should not be able to be interrupted by itself. Each one of the casts saves 6 bytes.*/
	((TIMER *)&Timer)->time.ms++;
	if(((TIMER *)&Timer)->time.ms == 19)
	{
		((TIMER *)&Timer)->time.ms = 0;
		((TIMER *)&Timer)->time.sec++;

		if(((TIMER *)&Timer)->time.sec == 60)
		{
			((TIMER *)&Timer)->time.sec = 0;
			((TIMER *)&Timer)->time.min++;
		}

		if(((TIMER *)&Timer)->print)
		{
			PrintTime();
		}
	}
}

void _main(void)
{
	/*a buffer for sprintf*/
	char buffer[50];
	/*a variable to save the current font in for restoring at the end of the program*/
	short oldFont;
	/*this must not change for saved games to work properly*/
	short menuRetVal = MAIN_MENU_START;
	/*you can set this one to anything you want*/
	short gameRetVal = GAME_LOSE;
	/*a pointer to the file that stores nearly everything*/
	DATA_FILE *file;
	/*stores the handle of the data file for later unlocking*/
	HANDLE fileHandle;
	/*places to save the interrupt handlers*/
	INT_HANDLER oldInt1;
	INT_HANDLER oldInt5;
	/*the pointer to all the dynamically allocated stuff*/
	MEM_BLOCK *master;

	/*load or create the data file*/
	file = LoadDataFile(&fileHandle);
	if(file == NULL)
	{
		/*if the file was not created, then you need memory for the file, the memory block, and
		grayscale*/
		sprintf(buffer, "Memory Low: Please Free %lu Bytes", sizeof(MEM_BLOCK) + sizeof(DATA_FILE) +
		(HW_VERSION == 1 ? GRAY_MEM_HW1 : GRAY_MEM_HW2));
		ST_helpMsg(buffer);
		return;
	}

	/*allocate all the memory except for grayscale that this game is going to need*/
	master = malloc(sizeof(MEM_BLOCK));
	/*if there was not enough memory, display a message and exit*/
	if(master == NULL)
	{
		CloseDataFile(file, fileHandle);
		sprintf(buffer, "Memory Low: Please Free %lu Bytes", sizeof(MEM_BLOCK) + (HW_VERSION == 1 ?
		GRAY_MEM_HW1 : GRAY_MEM_HW2));
		ST_helpMsg(buffer);
		return;
	}

	/*set up the initial position of the pointers for the autoclear queue*/
	master->queue.front = master->queue.back = 0;

	/*turn off int 1 & 5*/
	oldInt1 = GetIntVec(AUTO_INT_1);
	oldInt5 = GetIntVec(AUTO_INT_5);
	SetIntVec(AUTO_INT_1, DUMMY_HANDLER);
	SetIntVec(AUTO_INT_5, DUMMY_HANDLER);

	/*save the screen*/
	FastCopyScreen_R(LCD_MEM, master->bufferMain);
	/*set the font to 4x6 (proportional) and save what the font was for later restoring*/
	oldFont = FontSetSys(F_4x6);

	do
	{
		if(!file->data.savedGame)
		{
			/*if you are not playing a saved game, go to the main menu*/
			menuRetVal = MainMenu(file->data.mines, &file->data.settings, &file->data.highScores,
			master->bufferMainMenu);
		}

		/*if you did not quit out of the menu*/
		if(menuRetVal != MAIN_MENU_ESC)
		{
			if(!file->data.savedGame)
			{
				/*if you are not playing a saved game, set up various stuff for a new game*/
				file->data.state.flags = 0;
				file->data.state.emptySquaresCleared = 0;
				file->data.state.level = -1;
				file->data.state.nonZeroSquares = 0;
				file->data.state.squaresCleared = 0;
				file->data.state.score = 0;
				file->data.state.cursor = (LOCATION){0, 0};
			}

			do
			{
				if(!file->data.savedGame)
				{
					/*set up the size and number of mines for the minefield*/
					SetupGameParameters(file->data.mines, &file->data.settings, &file->data.state);
					/*put the mines and numbers in the minefield*/
					SetupMinefield(file->secretData.minefield, file->data.settings, &file->data.state);
				}
				else
				{
					/*saved game loaded, no need to load it again*/
					file->data.savedGame = FALSE;
					/*restore the time from the saved game*/
					memcpy((TIME *)&Timer.time, &file->data.time, sizeof(TIME));
				}

				/*start playing the game*/
				gameRetVal = PlayGame(&file->data.savedGame, file->secretData.minefield,
				file->data.settings, &file->data.state, master->backgroundSave, &master->queue,
				file->data.mines);
			} while((file->data.settings.scoring == SCORING_SCORED && gameRetVal != GAME_LOSE) &&
			gameRetVal != GAME_SAVE && gameRetVal != GAME_NO_GRAY);

			/*if you are playing a game type that has high scores*/
			if(file->data.settings.scoring == SCORING_SCORED || (file->data.settings.scoring ==
			SCORING_TIMED && file->data.settings.difficulty != DIFFICULTY_CUSTOM && gameRetVal ==
			GAME_WIN))
			{
				CheckHighScores(file->data.settings, &file->data.highScores, file->data.state.score);
			}
		}
	} while(menuRetVal != MAIN_MENU_ESC && gameRetVal != GAME_SAVE && gameRetVal != GAME_NO_GRAY);
	/*loop until [ESC] is pressed in the main menu or [clear] is pressed in game, or exit if the grayscale
	failed to turn on*/

	if(gameRetVal == GAME_SAVE)
	{
		/*if you are saving the game, save the time in the save game file*/
		memcpy(&file->data.time, (TIME *)&Timer.time, sizeof(TIME));
	}

	CloseDataFile(file, fileHandle);

	SetIntVec(AUTO_INT_1, oldInt1);
	SetIntVec(AUTO_INT_5, oldInt5);

	/*restore the screen before exiting*/
	FastCopyScreen_R(master->bufferMain, LCD_MEM);
	/*free the memory that this program used*/
	free(master);
	/*set the font back to whatever it was before the program started*/
	FontSetSys(oldFont);

	if(gameRetVal == GAME_NO_GRAY)
	{
		/*if grayscale did not turn on, then you need memory for grayscale*/
		sprintf(buffer, "Memory Low: Please Free %lu Bytes", (unsigned long)(HW_VERSION == 1 ?
		GRAY_MEM_HW1 : GRAY_MEM_HW2));
		ST_helpMsg(buffer);
	}
}

/*attempts to load the data file for this program. stores the handle of the file for later unlocking*/
static DATA_FILE *LoadDataFile(HANDLE *fileHandle)
{
	char hash[16];
	char *filep;
	unsigned short i;
	DATA_FILE *file = NULL;
	MD5_CTX context;
	SYM_ENTRY *file_vat;
	static FILE_SIG sig;

	/*this variable is used to verify that the file was created by this program, on the same type of
	calculator, with the same random number generator settings*/
	sig = (FILE_SIG){calc(0, 1), HW_VERSION, {'P', 'J', 'R'}, RANDOM_STRENGTH};

	/*try to find the data file*/
	file_vat = SymFindPtr(DATA_NAME, 0);

	/*if the file was found*/
	if(file_vat != NULL)
	{
		/*and it is archived*/
		if(file_vat->flags.bits.archived)
		{
			/*try to unarchive it*/
			if(!EM_moveSymFromExtMem(DATA_NAME, HS_NULL))
			{
				/*if there wasn't enough memory to unarchive it*/
				return NULL;
			}
		}
		/*get a pointer to the data for the next set of checks and store the handle so that I can unlock
		the file later*/
		file = HLock(*fileHandle = file_vat->handle);

		/*hash the file to check for corruption. do not hash the footer portion of the file becuase that
		is where the hash to compare to is stored*/
		MD5Init(&context);
		MD5Update(&context, (char *)&file->header, sizeof(file->header));
		MD5Update(&context, (char *)&file->data, sizeof(file->data));
		MD5Update(&context, (char *)&file->secretData, sizeof(file->secretData));
		MD5Final(hash, &context);
	}

	if(file_vat == NULL || memcmp(&file->header.sig, &sig, sizeof(FILE_SIG)) || memcmp(file->footer.hash,
	hash, 16))
	{
		/*the file is either nonexistent, corrupted, or not created by this program if the program gets to
		this point*/

		randomize_long();

		/*if the file is corrupted*/
		if(file_vat != NULL)
		{
			/*delete it*/
			SymDel(DATA_NAME);
		}

		/*create the file*/
		file_vat = DerefSym(SymAdd(DATA_NAME));
		file_vat->handle = HeapAlloc(sizeof(DATA_FILE));

		/*if there was not enough memory to create the file*/
		if(file_vat->handle == H_NULL)
		{
			return NULL;
		}

		/*get a pointer to the file, lock the handle, and save it for later unlocking*/
		file = HLock(*fileHandle = file_vat->handle);

		/*write in the default file contents*/
		file->header.size = sizeof(DATA_FILE) - 2;
		file->header.sig = sig;

		file->data.savedGame = 0;
		memcpy(file->data.mines, (unsigned char[]){30, 30, 30}, 3 * sizeof(char));

		file->data.settings.row = 0;
		file->data.settings.column = 0;
		file->data.settings.delay = 100;
		file->data.settings.misc = 0;
		file->data.settings.difficulty = DIFFICULTY_PROGRESSIVE;
		file->data.settings.numbering = NUMBERING_TORUS;
		file->data.settings.scoring = SCORING_NONE;
		file->data.settings.movement = MOVEMENT_FLYING;
		file->data.settings.height = 1;
		file->data.settings.width = 1;
		file->data.settings.mines = 0;
		file->data.settings.hw = sig.hw;

		file->data.state.flags = 0;
		file->data.state.level = -1;
		file->data.state.squaresCleared = 0;
		file->data.state.score = 0;
		file->data.state.cursor = (LOCATION){0, 0};

		file->data.highScores.numbering = NUMBERING_PLANE;
		file->data.highScores.row = 0;
		file->data.highScores.scored = HIGHSCORES_TIMED;
		file->data.highScores.movement = MOVEMENT_WALKING;
		memset(file->data.highScores.scores, 0, 2 * sizeof(unsigned long));
		memcpy(file->data.highScores.times, (TIME[]){{127, 63, 31}, {127, 63, 31}, {127, 63, 31},
		{127, 63, 31}, {127, 63, 31}, {127, 63, 31}, {127, 63, 31}, {127, 63, 31}, {127, 63, 31}},
		9 * sizeof(TIME));

		SaveRandState(&file->data.randomState);

		memcpy(file->footer.type, (char []){0, 'D', 'a', 't', 'a', 0, OTH_TAG}, 7);
	}
	else
	{
		/*the file is valid if this is executed*/

		/*restore the state of the random number generator that was used to scramble the secret data
		portion*/
		srand_long(file->data.seed);

		/*unscramble the secret data*/
		for(i = 0, filep = (char *)&file->secretData; i < sizeof(file->secretData); i++)
		{
			*filep++ ^= random_long(256);
		}

		/*restore the random state that was saved from gameplay*/
		LoadRandState(&file->data.randomState);
	}

	return file;
}

/*closes down the data file and unlocks the file handle*/
static void CloseDataFile(DATA_FILE *file, HANDLE fileHandle)
{
	/*the first two varibales are used for obscuring the data about the minefield*/
	char *filep;
	unsigned short i;
	MD5_CTX context;

	/*save the state of the random number generator so that it does not get messed up by the obscuring
	process*/
	SaveRandState(&file->data.randomState);

	/*save a random number and seed the generator with it. this number is used to seed the generator again
	when the file is loaded*/
	srand_long(file->data.seed = rand_long());

	/*obscure the data about he state of the minefield*/
	for(i = 0, filep = (char *)&file->secretData; i < sizeof(file->secretData); i++)
	{
		*filep++ ^= random_long(256);
	}

	/*hash the file to gaurd against modification*/
	MD5Init(&context);
	MD5Update(&context, (char *)&file->header, sizeof(file->header));
	MD5Update(&context, (char *)&file->data, sizeof(file->data));
	MD5Update(&context, (char *)&file->secretData, sizeof(file->secretData));
	MD5Final(file->footer.hash, &context);

	/*unlock the file's handle*/
	HeapUnlock(fileHandle);

	/*archive the file*/
	EM_moveSymToExtMem(DATA_NAME, HS_NULL);
}

/*the first menu that the user sees. allows the user to select the type of game they wish to play, as well
as veiwing high scores and adjusting the key timing*/
static MAIN_MENU_RETURN MainMenu(unsigned char mines[], GAME_SETTINGS *settings, HIGH_SCORES *highScores,
								 LCD_BUFFER buffer)
{
	/*a variable to use for rowread*/
	unsigned short key;

	ClrScr();
	MainMenuText(mines, *settings);
	MainMenuHighlights(5, mines, (char []){0, 1, 2, 3, 4}, *settings);

	do
	{
		key = _rowread(0xFFFE);

		/*up or down pressed*/
		if(key & calc(5, 160))
		{
			/*erase the dots that indicate the current row*/
			MainMenuDrawDot(*settings);
		}

		/*up pressed*/
		if(key & calc(1, 32))
		{
			/*wrap the row counter over the top*/
			if(settings->row == 0)
			{
				settings->row = 5;
			}
			else
			{
				settings->row--;

				/*skip the difficulty row if you are playing a scored game*/
				if(settings->row == 2 && settings->scoring == SCORING_SCORED)
				{
					settings->row--;
				}
			}

			/*wait until up released*/
			while(_rowread(0xFFFE));
		}

		/*down pressed*/
		if(key & calc(4, 128))
		{
			/*wrap the row counter over the bottom*/
			if(settings->row == 5)
			{
				settings->row = 0;
			}
			else
			{
				settings->row++;

				/*skip the difficulty row if you are playing a scored game*/
				if(settings->row == 2 && settings->scoring == SCORING_SCORED)
				{
					settings->row++;
				}
			}

			/*wait until down released*/
			while(_rowread(0xFFFE));
		}

		/*up or down pressed*/
		if(key & calc(5, 160))
		{
			/*redraw the dots that indicate the current row*/
			MainMenuDrawDot(*settings);
		}

		/*left or right pressed*/
		if(key & calc(10, 80))
		{
			/*if you are not on the start game row*/
			if(settings->row < 5)
			{
				/*erase the highlight that indicates the current setting of the row*/
				MainMenuDrawHighlight(mines, *settings);
			}
		}

		/*left pressed*/
		if(key & calc(2, 16))
		{
			switch(settings->row)
			{
				/*movement row*/
				case 0:
				if(settings->movement == MOVEMENT_FLYING)
				{
					/*erase the number of mines display for progressive*/
					MainMenuDrawProgressive(A_REVERSE, mines, *settings);

					settings->movement = MOVEMENT_WALKING;

					/*walking and torus is an invlaid combination*/
					MainMenuHighlights(1, mines, (char []){1}, *settings);
					settings->numbering = NUMBERING_PLANE;
					MainMenuHighlights(1, mines, (char []){1}, *settings);

					/*redraw and highlight the mines display for progressive*/
					MainMenuDrawProgressive(A_XOR, mines, *settings);

					/*walking and scored mode is an invalid combination*/
					if(settings->scoring == SCORING_SCORED)
					{
						MainMenuHighlights(3, mines, (char []){1, 3, 4}, *settings);
						settings->scoring = SCORING_NONE;
						/*ungray the difficulty row*/
						MainMenuText(mines, *settings);
						MainMenuHighlights(4, mines, (char []){1, 2, 3, 4}, *settings);
					}
				}
				else
				{
					MainMenuDrawProgressive(A_REVERSE, mines, *settings);
					settings->movement = MOVEMENT_FLYING;
					MainMenuDrawProgressive(A_XOR, mines, *settings);
				}
				break;

				/*numbering row*/
				case 1:
				if(settings->numbering == NUMBERING_TORUS)
				{
					MainMenuDrawProgressive(A_REVERSE, mines, *settings);
					settings->numbering = NUMBERING_PLANE;
					MainMenuDrawProgressive(A_XOR, mines, *settings);
				}
				else
				{
					MainMenuDrawProgressive(A_REVERSE, mines, *settings);

					settings->numbering = NUMBERING_TORUS;

					MainMenuHighlights(1, mines, (char []){0}, *settings);
					/*torus and walking is an invalid combination*/
					settings->movement = MOVEMENT_FLYING;
					MainMenuHighlights(1, mines, (char []){0}, *settings);

					MainMenuDrawProgressive(A_XOR, mines, *settings);
				}
				break;

				/*difficulty row*/
				case 2:
				if(settings->difficulty == DIFFICULTY_BEGINNER)
				{
					settings->difficulty = DIFFICULTY_PROGRESSIVE;

					MainMenuHighlights(1, mines, (char []){3}, *settings);
					/*progressive mode cannot be scored or timed*/
					settings->scoring = SCORING_NONE;
					MainMenuHighlights(1, mines, (char []){3}, *settings);
				}
				else
				{
					settings->difficulty--;
				}
				break;

				/*scoring row*/
				case 3:
				if(settings->scoring == SCORING_NONE)
				{
					/*hide difficulty*/
					MainMenuHighlights(4, mines, (char []){0, 1, 2, 4}, *settings);
					settings->scoring = SCORING_SCORED;
					MainMenuText(mines, *settings);
					/*scored mode and wolking is an invalid combination*/
					settings->movement = MOVEMENT_FLYING;
					MainMenuHighlights(3, mines, (char []){0, 1, 4}, *settings);
				}
				else
				{
					if(settings->scoring == SCORING_SCORED)
					{
						/*show difficulty*/
						MainMenuHighlights(3, mines, (char []){0, 1, 4}, *settings);
						settings->scoring = SCORING_TIMED;
						MainMenuText(mines, *settings);
						MainMenuHighlights(4, mines, (char []){0, 1, 2, 4}, *settings);
					}
					else
					{
						settings->scoring--;
					}

					/*progressive and any scored mode is an invalid combination*/
					if(settings->scoring != SCORING_NONE && settings->difficulty ==
					DIFFICULTY_PROGRESSIVE)
					{
						MainMenuHighlights(1, mines, (char []){2}, *settings);
						settings->difficulty = DIFFICULTY_BEGINNER;
						MainMenuHighlights(1, mines, (char []){2}, *settings);
					}
				}
				break;

				/*the misc row. I stuck everything that didn't fit anywhere else here. this row has high
				score viewing, custom game setup, and adjusting the key timing*/
				case 4:
				if(settings->misc == 0)
				{
					settings->misc = 2;
				}
				else
				{
					settings->misc--;
				}
			}
			/*wait until left released*/
			while(_rowread(0xFFFE));
		}

		/*right pressed*/
		if(key & calc(8, 64))
		{
			switch(settings->row)
			{
				/*movement row*/
				case 0:
				if(settings->movement == MOVEMENT_FLYING)
				{
					MainMenuDrawProgressive(A_REVERSE, mines, *settings);

					settings->movement = MOVEMENT_WALKING;

					MainMenuHighlights(1, mines, (char []){1}, *settings);
					/*walking and toroidal numbering is an invalid combination*/
					settings->numbering = NUMBERING_PLANE;
					MainMenuHighlights(1, mines, (char []){1}, *settings);

					MainMenuDrawProgressive(A_XOR, mines, *settings);

					/*walking and scored mode is an invalid combination*/
					if(settings->scoring == SCORING_SCORED)
					{
						/*show the difficulty row*/
						MainMenuHighlights(3, mines, (char []){1, 3, 4}, *settings);
						settings->scoring = SCORING_NONE;
						MainMenuText(mines, *settings);
						MainMenuHighlights(4, mines, (char []){1, 2, 3, 4}, *settings);
					}
				}
				else
				{
					MainMenuDrawProgressive(A_REVERSE, mines, *settings);
					settings->movement = MOVEMENT_FLYING;
					MainMenuDrawProgressive(A_XOR, mines, *settings);
				}
				break;

				/*numbering row*/
				case 1:
				if(settings->numbering == NUMBERING_TORUS)
				{
					MainMenuDrawProgressive(A_REVERSE, mines, *settings);
					settings->numbering = NUMBERING_PLANE;
					MainMenuDrawProgressive(A_XOR, mines, *settings);
				}
				else
				{
					MainMenuDrawProgressive(A_REVERSE, mines, *settings);

					settings->numbering = NUMBERING_TORUS;

					MainMenuHighlights(1, mines, (char []){0}, *settings);
					/*toroidal numbering and walking is an invalid combination*/
					settings->movement = MOVEMENT_FLYING;
					MainMenuHighlights(1, mines, (char []){0}, *settings);

					MainMenuDrawProgressive(A_XOR, mines, *settings);
				}
				break;

				/*difficulty row*/
				case 2:
				if(settings->difficulty == DIFFICULTY_PROGRESSIVE)
				{
					settings->difficulty = DIFFICULTY_BEGINNER;
				}
				else
				{
					settings->difficulty++;

					/*progressive and any type of scoring is an invalid combination*/
					if(settings->difficulty == DIFFICULTY_PROGRESSIVE)
					{
						MainMenuHighlights(1, mines, (char []){3}, *settings);
						settings->scoring = SCORING_NONE;
						MainMenuHighlights(1, mines, (char []){3}, *settings);
					}
				}
				break;

				/*scoring row*/
				case 3:
				if(settings->scoring == SCORING_SCORED)
				{
					/*show difficulty*/
					MainMenuHighlights(3, mines, (char []){0, 1, 4}, *settings);
					settings->scoring = SCORING_NONE;
					MainMenuText(mines, *settings);
					MainMenuHighlights(4, mines, (char []){0, 1, 2, 4}, *settings);
				}
				else
				{
					if(settings->scoring == SCORING_TIMED)
					{
						/*hide difficulty*/
						MainMenuHighlights(4, mines, (char []){0, 1, 2, 4}, *settings);
						settings->scoring = SCORING_SCORED;
						MainMenuText(mines, *settings);
						/*scored mode and walking is an invalid combination*/
						settings->movement = MOVEMENT_FLYING;
						MainMenuHighlights(3, mines, (char []){0, 1, 4}, *settings);
					}
					else
					{
						settings->scoring++;
					}

					/*progressive and any kind of scoring is an invalid combination*/
					if(settings->scoring != SCORING_NONE && settings->difficulty ==
					DIFFICULTY_PROGRESSIVE)
					{
						MainMenuHighlights(1, mines, (char []){2}, *settings);
						settings->difficulty = DIFFICULTY_BEGINNER;
						MainMenuHighlights(1, mines, (char []){2}, *settings);
					}
				}
				break;

				case 4:
				if(settings->misc == 2)
				{
					settings->misc = 0;
				}
				else
				{
					settings->misc++;
				}
			}
			/*wait until right released*/
			while(_rowread(0xFFFE));
		}

		/*left or right pressed*/
		if(key & calc(10, 80))
		{
			/*if you are not on the game start row*/
			if(settings->row < 5)
			{
				MainMenuDrawHighlight(mines, *settings);
			}
		}

		/*second or hand*/
		if(key & calc(16, 8))
		{
			/*wait until second or hand is released*/
			while(_rowread(0xFFFE));
			switch(settings->row)
			{
				/*misc row*/
				case 4:
				switch(settings->misc)
				{
					case 0:
					/*save the screen*/
					FastCopyScreen_R(LCD_MEM, buffer);
					HighScoreMenu(highScores);
					/*restore the screen*/
					FastCopyScreen_R(buffer, LCD_MEM);
					break;

					case 1:
					sprintf(buffer, "Custom: %ux%ux%u", settings->height, settings->width,
					settings->mines);
					/*erase the custom text*/
					FastFilledRect_Erase_R(LCD_MEM, 0, 58, DrawStrWidth(buffer, F_4x6), 64);
					/*save the screen*/
					FastCopyScreen_R(LCD_MEM, buffer);
					CustomMenu(settings);
					/*restore the screen*/
					FastCopyScreen_R(buffer, LCD_MEM);
					/*redraw the custom text*/
					sprintf(buffer, "Custom: %ux%ux%u", settings->height, settings->width,
					settings->mines);
					DrawStr(0, 59, buffer, settings->scoring == SCORING_SCORED ? A_SHADED : A_NORMAL);
					/*highlight the custom text if necessary*/
					if(settings->difficulty == DIFFICULTY_CUSTOM && settings->scoring != SCORING_SCORED)
					{
						FastFilledRect_Invert_R(LCD_MEM, 0, 58, DrawStrWidth(buffer, F_4x6), 64);
					}
					break;

					case 2:
					/*save the screen*/
					FastCopyScreen_R(LCD_MEM, buffer);
					KeyDelayMenu(settings);
					/*restore the screen*/
					FastCopyScreen_R(buffer, LCD_MEM);
					break;
				}
				break;

				case 5:
				/*start a game*/
				return MAIN_MENU_START;
			}
		}
		/*loop until [ESC] pressed*/
	} while(!(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)));
	/*wait until [ESC] released*/
	while(_rowread(calc(0xFFBF, 0xFEFF)));

	/*quit the program*/
	return MAIN_MENU_ESC;
}

/*draws all the text for the main menu. takes as parameters the number of mines for progressive and the
current game settings*/
static void MainMenuText(unsigned char mines[], GAME_SETTINGS settings)
{
	/*a buffer for sprintf*/
	char buffer[50];

	sprintf(buffer, "Ultimate Minesweeper 2 r%u", RELEASE);
	DrawStr(0, 0, buffer, A_REPLACE);
	DrawStr(calc(117, 195), 3, "Use \x15\x16\x17\x18", A_REPLACE);
	DrawStr((DrawStrWidth(buffer, F_4x6) - 57) / 2, 6, "By Peter J. Rowe", A_REPLACE);
	DrawStr(calc(106, 182), 9, calc("+ [2nd] or [ESC]", "+ [Hand] or [ESC]"), A_REPLACE);
	DrawStr((DrawStrWidth(buffer, F_4x6) - 62) / 2, 12, "mig53@yahoo.com", A_REPLACE);
	DrawStr(calc(62, 102), 21, "Movement", A_REPLACE);
	DrawStr(0, 27, "Walking", A_REPLACE);
	DrawStr(calc(138, 218), 27, "Flying", A_REPLACE);
	DrawStr(calc(61, 101), 33, "Numbering", A_REPLACE);
	DrawStr(0, 39, "Plane", A_REPLACE);
	DrawStr(calc(140, 220), 39, "Torus", A_REPLACE);
	DrawStr(calc(64, 104), 45, "Difficulty", (settings.scoring == 2 ? A_SHADED : A_REPLACE));
	DrawStr(0, 52, "Beginner", (settings.scoring == 2 ? A_SHADED : A_REPLACE));
	DrawStr(calc(58, 98), 52, "Intermediate", (settings.scoring == 2 ? A_SHADED : A_REPLACE));
	DrawStr(calc(127, 207), 52, "Advanced", (settings.scoring == 2 ? A_SHADED : A_REPLACE));
	sprintf(buffer, "Custom: %ux%ux%u", settings.height, settings.width, settings.mines);
	DrawStr(0, 59, buffer, (settings.scoring == 2 ? A_SHADED : A_REPLACE));
	sprintf(buffer, "    Progressive: %u mines", mines[settings.movement + settings.numbering] / 2);
	DrawStr(calc(159, 239) - DrawStrWidth(buffer, F_4x6), 59, buffer, (settings.scoring == 2 ? A_SHADED :
	A_REPLACE));
	DrawStr(calc(49, 89), 66, "Evaluation System", A_REPLACE);
	DrawStr(0, 73, "None", A_REPLACE);
	DrawStr(calc(70, 110), 73, "Timed", A_REPLACE);
	DrawStr(calc(135, 215), 73, "Scored", A_REPLACE);
	DrawStr(calc(56, 96), 80, "Miscellaneous", A_REPLACE);
	DrawStr(0, 87, "High Scores", A_REPLACE);
	DrawStr(calc(57, 97), 87, "Custom Setup", A_REPLACE);
	DrawStr(calc(122, 202), 87, "Key Timing", A_REPLACE);
	DrawStr(calc(60, 100), calc(95, 123), "Start Game", A_REPLACE);
}

/*draws the highlights for the current settings of each row. takes as parameters the number of rows to
work on, the number of mines for progressive, the rows to work on, and the current game settings*/
static void MainMenuHighlights(unsigned char numRow, unsigned char mines[], unsigned char rows[],
GAME_SETTINGS settings)
{
	short i;
	/*make a copy of the settings so that I can modify it without messing stuff up*/
	GAME_SETTINGS tempSettings = settings;

	/*make sure that the dots that indicate the row are drawn*/
	MainMenuDrawDot(settings);

	/*iterate through the rows*/
	for(i = 0; i < numRow; i++)
	{
		if(rows[i] != 2 || settings.scoring != SCORING_SCORED)
		{
			/*set the current row to be the appropriate entry from the rows argument*/
			tempSettings.row = rows[i];

			/*draw the highlight for the current row*/
			MainMenuDrawHighlight(mines, tempSettings);
		}
	}
}

/*draws the display of the number of mines for progressive*/
static void MainMenuDrawProgressive(char attr, char mines[], GAME_SETTINGS settings)
{
	/*a buffer for sprintf*/
	char buffer[50];

	sprintf(buffer, "Progressive: %u mines", mines[settings.movement + settings.numbering] / 2);
	if(attr == A_XOR)
	{
		DrawStr(calc(159, 239) - DrawStrWidth(buffer, F_4x6), 59, buffer, (settings.scoring ==
		SCORING_SCORED ? A_SHADED : A_NORMAL));
	}
	if((settings.difficulty == DIFFICULTY_PROGRESSIVE || attr == A_REVERSE) && settings.scoring !=
	SCORING_SCORED)
	{
		FastFillRect_R(LCD_MEM, calc(159, 239) - DrawStrWidth(buffer, F_4x6), 58, calc(159, 239), 64,
		attr);
	}
}

/*draws the dots that indicate the current row. parameter is the current game settings that include the
current row*/
static void MainMenuDrawDot(GAME_SETTINGS settings)
{
	static unsigned char dot[] = {0xE0, 0xE0, 0xE0};
	static unsigned char dotX[][2] = {{calc(59, 99), calc(98, 138)}, {calc(58, 98), calc(99, 139)},
	{calc(61, 101), calc(97, 137)}, {calc(46, 86), calc(112, 152)}, {calc(53, 93), calc(105, 145)},
	{calc(57, 97), calc(100, 140)}};
	static unsigned char dotY[] = {22, 34, 46, 67, 81, calc(96, 124)};

	/*draw the left dot*/
	Sprite8_XOR_R(dotX[settings.row][0], dotY[settings.row], 3, dot, LCD_MEM);
	/*draw the right dot*/
	Sprite8_XOR_R(dotX[settings.row][1], dotY[settings.row], 3, dot, LCD_MEM);
}

/*draws the highlight for the current setting of the current row. parameters are the number of mines for
progressive and the current game settings*/
static void MainMenuDrawHighlight(unsigned char mines[], GAME_SETTINGS settings)
{
	/*a buffer for sprintf*/
	char buffer[50];
	/*the setting of the current row*/
	unsigned char rowSetting;
	/*the locations that need highlighting*/
	static SCR_RECT boxes[][3] = {{{{calc(138, 218), 26, calc(159, 239), 32}}, {{0, 26, 27, 32}}},
	{{{calc(140, 220), 38, calc(159, 239), 44}}, {{0, 38, 19, 44}}}, {{{0, 51, 30, 57}}, {{calc(58, 98),
	51, calc(102, 142), 57}}, {{calc(127, 207), 51, calc(159, 239), 57}}}, {{{0, 72, 17, 78}}, {{calc(70,
	110), 72, calc(90, 130), 78}}, {{calc(135, 215), 72, calc(159, 239), 78}}}, {{{0, 86, 39, 92}},
	{{calc(57, 97), 86, calc(102, 142), 92}}, {{calc(122, 202), 86, calc(159, 239), 92}}}};

	/*get the setting of the current row*/
	rowSetting = (unsigned char[]){settings.movement, settings.numbering, settings.difficulty,
	settings.scoring, settings.misc}[settings.row];

	switch(rowSetting)
	{
		/*custom is variable width, so it has to be handled in a seperate case*/
		case DIFFICULTY_CUSTOM:
		sprintf(buffer, "Custom: %ux%ux%u", settings.height, settings.width, settings.mines);
		FastFilledRect_Invert_R(LCD_MEM, 0, 58, DrawStrWidth(buffer, F_4x6), 64);
		break;

		/*progressive is also variable width*/
		case DIFFICULTY_PROGRESSIVE:
		sprintf(buffer, "Progressive: %u mines", mines[settings.movement + settings.numbering] / 2);
		FastFilledRect_Invert_R(LCD_MEM, calc(159, 239) - DrawStrWidth(buffer, F_4x6), 58, calc(159, 239),
		64);
		break;

		default:
		FastFilledRect_Invert_R(LCD_MEM, boxes[settings.row][rowSetting].xy.x0,
		boxes[settings.row][rowSetting].xy.y0, boxes[settings.row][rowSetting].xy.x1,
		boxes[settings.row][rowSetting].xy.y1);
	}
}

/*displays the high scores. the parameter is the high score structure*/
static void HighScoreMenu(HIGH_SCORES *scores)
{
	/*a variable for use in rowread*/
	unsigned short key;

	ClrScr();
	HighScoreText(*scores);
	HighScoreHighlights(*scores);
	HighScoreDrawDot(*scores);
	HighScoreDrawScores(*scores);

	do
	{
		key = _rowread(0xFFFE);

		/*up or down pressed*/
		if(key & calc(5, 160))
		{
			/*erase the dot that indicates the current row*/
			HighScoreDrawDot(*scores);
		}

		/*up pressed*/
		if(key & calc(1, 32))
		{
			if(scores->row == 0)
			{
				/*there are only two rows if you are looking at scores, and three if you are looking at
				times*/
				scores->row = (scores->scored == 1 ? 1 : 2);
			}
			else
			{
				scores->row--;
			}
			/*wait for up to be released*/
			while(_rowread(0xFFFE));
		}

		/*down pressed*/
		if(key & calc(4, 128))
		{
			/*there are only two rows if you are looking at scores, and three if you are looking at
			times*/
			if(scores->row == (scores->scored == 1 ? 1 : 2))
			{
				scores->row = 0;
			}
			else
			{
				scores->row++;
			}
			/*wait for down to be released*/
			while(_rowread(0xFFFE));
		}

		/*up or down pressed*/
		if(key & calc(5, 160))
		{
			/*resotre the dots that indicate the current row*/
			HighScoreDrawDot(*scores);
		}

		/*left or right pressed*/
		if(key & calc(10, 80))
		{
			/*erase the highlight on the current row*/
			HighScoreHighlights(*scores);

			switch(scores->row)
			{
				case 0:
				/*erase the area that the scores are drawn in and the last row*/
				FastFilledRect_Erase_R(LCD_MEM, 0, 24, calc(159, 239), 77);
				if(scores->scored == HIGHSCORES_SCORED)
				{
					scores->scored = HIGHSCORES_TIMED;
					HighScoreText(*scores);
				}
				else
				{
					scores->scored = HIGHSCORES_SCORED;
					/*scored and walking is an invalid combination*/
					scores->movement = MOVEMENT_FLYING;
				}
				break;

				case 1:
				if(scores->movement == MOVEMENT_WALKING)
				{
					scores->movement = MOVEMENT_FLYING;
					/*erase the area where the times are displayed*/
					FastFilledRect_Erase_R(LCD_MEM, 0, 36, calc(159, 239), 77);
				}
				else
				{
					scores->movement = MOVEMENT_WALKING;
					/*walking and toroidal numbering is an invalid combination*/
					scores->numbering = NUMBERING_PLANE;
					if(scores->scored == HIGHSCORES_SCORED)
					{
						/*erase the area that the scores are drawn in and the last row*/
						FastFilledRect_Erase_R(LCD_MEM, 0, 24, calc(159, 239), 77);
						scores->scored = HIGHSCORES_TIMED;
						HighScoreText(*scores);
					}
					else
					{
						/*erase the area where the times are displayed*/
						FastFilledRect_Erase_R(LCD_MEM, 0, 36, calc(159, 239), 77);
					}
				}
				break;

				case 2:
				if(scores->numbering == NUMBERING_PLANE)
				{
					scores->numbering = NUMBERING_TORUS;
					/*toroidal numbering and walking is an invalid combination*/
					scores->movement = MOVEMENT_FLYING;
					/*erase the area where the times are displayed*/
					FastFilledRect_Erase_R(LCD_MEM, 0, 36, calc(159, 239), 77);
				}
				else
				{
					scores->numbering = NUMBERING_PLANE;
					/*erase the area where the times are displayed*/
					FastFilledRect_Erase_R(LCD_MEM, 0, 36, calc(159, 239), 77);
				}
			}

			HighScoreHighlights(*scores);
			HighScoreDrawScores(*scores);

			/*wait for left and right to be released*/
			while(_rowread(0xFFFE));
		}
		/*loop until [ESC] pressed*/
	} while(!(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)));
	/*wait for [ESC] to be released*/
	while(_rowread(calc(0xFFBF, 0xFEFF)));
}

/*darws the text for the high score menu. the parameter is the current settings of the high score menu*/
static void HighScoreText(HIGH_SCORES scores)
{
	DrawStr(calc(53, 93), 0, "Scores or Times", A_NORMAL);
	DrawStr(0, 6, "Scores", A_NORMAL);
	DrawStr(calc(140, 220), 6, "Times", A_NORMAL);
	DrawStr(calc(62, 102), 12, "Play Style", A_NORMAL);
	DrawStr(0, 18, "Walking", A_NORMAL);
	DrawStr(calc(138, 218), 18, "Flying", A_NORMAL);
	if(scores.scored != 1)
	{
		DrawStr(calc(61, 101), 24, "Numbering", A_NORMAL);
		DrawStr(0, 30, "Plane", A_NORMAL);
		DrawStr(calc(140, 220), 30, "Torus", A_NORMAL);
	}
	DrawStr(calc(51, 91), calc(95, 123), "Use \x15\x16\x17\x18 + [ESC]", A_NORMAL);
}

/*highlights the current settings of all rows*/
static void HighScoreHighlights(HIGH_SCORES scores)
{
	short i;
	/*make a copy of the high score structure so that it can be modified without messing anything up*/
	HIGH_SCORES tempScores = scores;

	for(i = 0; i < (scores.scored == HIGHSCORES_SCORED ? 2 : 3); i++)
	{
		tempScores.row = i;
		HighScoreDrawHighlight(tempScores);
	}
}

/*draws the highlight for the current setting of the current row*/
static void HighScoreDrawHighlight(HIGH_SCORES scores)
{
	unsigned char rowSetting;
	static SCR_RECT boxes[][2] = {
	{{{calc(140, 220), 5, calc(159, 239), 11}}, {{0, 5, 23, 11}}},
	{{{calc(138, 218), 17, calc(159, 239), 23}}, {{0, 17, 27, 23}}},
	{{{calc(140, 220), 29, calc(159, 239), 35}}, {{0, 29, 19, 35}}}};

	/*get the setting of the current row*/
	rowSetting = (unsigned char[]){scores.scored, scores.movement, scores.numbering}[scores.row];

	FastFilledRect_Invert_R(LCD_MEM, boxes[scores.row][rowSetting].xy.x0,
	boxes[scores.row][rowSetting].xy.y0, boxes[scores.row][rowSetting].xy.x1,
	boxes[scores.row][rowSetting].xy.y1);
}

/*draws the dot that indicates the current row*/
static void HighScoreDrawDot(HIGH_SCORES scores)
{
	static unsigned char dot[] = {0xE0, 0xE0, 0xE0};
	static unsigned char dotX[][2] = {{calc(50, 90), calc(108, 148)}, {calc(59, 99), calc(98, 138)},
	{calc(58, 98), calc(99, 139)}};
	static unsigned char dotY[] = {1, 13, 25};

	/*draw the left dot*/
	Sprite8_XOR_R(dotX[scores.row][0], dotY[scores.row], 3, dot, LCD_MEM);
	/*draw the right dot*/
	Sprite8_XOR_R(dotX[scores.row][1], dotY[scores.row], 3, dot, LCD_MEM);
}

/*draws the scores or times, depending on the current settings*/
static void HighScoreDrawScores(HIGH_SCORES scores)
{
	const char movement[][6] = {"Plane", "Torus"};
	const char difficulty[][13] = {"Beginner", "Intermediate", "Advanced"};
	/*a buffer for sprintf*/
	char buffer[50];
	short i;

	if(scores.scored)
	{
		DrawStr(calc(60, 100), 24, "High Scores", A_NORMAL);
		for(i = 0; i < 2; i++)
		{
			DrawStr(0, 30 + 12 * i, movement[i], A_NORMAL);
			sprintf(buffer, "%lu", scores.scores[1 ^ i]);
			DrawStr(0, 36 + 12 * i, buffer, A_NORMAL);
		}
	}
	else
	{
		DrawStr(calc(62, 102), 36, "Best Times", A_NORMAL);

		for(i = 0; i < 3; i++)
		{
			DrawStr(0, 42 + 12 * i, difficulty[i], A_NORMAL);
			sprintf(buffer, "%02u:%02u", scores.times[scores.movement + scores.numbering][i].min,
			scores.times[scores.movement + scores.numbering][i].sec);
			DrawStr(0, 48 + 12 * i, buffer, A_NORMAL);
		}
	}
}

/*does the menu for setting up custom games. the parameter is the current game settings*/
static void CustomMenu(GAME_SETTINGS *settings)
{
	/*a variable to use with rowread*/
	unsigned short key;

	ClrScr();

	CustomMenuText(*settings);
	CustomMenuHighlight(*settings);

	do
	{
		key = _rowread(0xFFFE);

		/*if an arrow key pressed*/
		if(key & calc(15, 240))
		{
			/*erase the highlight on the currently selected setting*/
			CustomMenuHighlight(*settings);
		}

		/*up*/
		if(key & calc(1, 32))
		{
			switch(settings->column)
			{
				case 0:
				/*wrap the height*/
				if(settings->height == MAX_HEIGHT)
				{
					settings->height = 1;
					/*the mine density shall be a maximum of 0.5*/
					settings->mines = min((unsigned)(settings->width * settings->height / 2),
					settings->mines);
				}
				else
				{
					settings->height++;
				}
				break;

				case 1:
				/*wrap the width*/
				if(settings->width == MAX_WIDTH)
				{
					settings->width = 1;
					/*the mine density shall be a maximum of 0.5*/
					settings->mines = min((unsigned)(settings->width * settings->height / 2),
					settings->mines);
				}
				else
				{
					settings->width++;
				}
				break;

				case 2:
				/*wrap the mines at the max density of 0.5*/
				if(settings->mines == (unsigned)(settings->width * settings->height / 2))
				{
					settings->mines = 0;
				}
				else
				{
					settings->mines++;
				}
			}
			/*wait for up to be released*/
			while(_rowread(0xFFFE));
		}

		/*left*/
		if(key & calc(2, 16))
		{
			if(settings->column == 0)
			{
				settings->column = 2;
			}
			else
			{
				settings->column--;
			}
			/*wait for left to be released*/
			while(_rowread(0xFFFE));
		}

		/*down*/
		if(key & calc(4, 128))
		{
			switch(settings->column)
			{
				case 0:
				/*wrap the height*/
				if(settings->height == 1)
				{
					settings->height = MAX_HEIGHT;
				}
				else
				{
					settings->height--;
					/*the mine density shall be a maximum of 0.5*/
					settings->mines = min((unsigned)(settings->width * settings->height / 2),
					settings->mines);
				}
				break;

				case 1:
				/*wrap the width*/
				if(settings->width == 1)
				{
					settings->width = MAX_WIDTH;
				}
				else
				{
					settings->width--;
					/*the mine density shall be a maximum of 0.5*/
					settings->mines = min((unsigned)(settings->width * settings->height / 2),
					settings->mines);
				}
				break;

				case 2:
				/*wrap the mines at the max density of 0.5*/
				if(settings->mines == 0)
				{
					settings->mines = settings->width * settings->height / 2;
				}
				else
				{
					settings->mines--;
				}
			}
			/*wait for down to be released*/
			while(_rowread(0xFFFE));
		}

		/*right*/
		if(key & calc(8, 64))
		{
			if(settings->column == 2)
			{
				settings->column = 0;
			}
			else
			{
				settings->column++;
			}
			/*wait for right to be released*/
			while(_rowread(0xFFFE));
		}

		/*up or down*/
		if(key & calc(5, 160))
		{
			/*redraw the text for the menu*/
			CustomMenuText(*settings);
		}

		/*if an arrow key pressed*/
		if(key & calc(15, 240))
		{
			/*redraw the highlight on the curently selected setting*/
			CustomMenuHighlight(*settings);
		}
		/*loop until [ESC] pressed*/
	} while(!(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)));
	/*wait unitl [ESC] released*/
	while(_rowread(calc(0xFFBF, 0xFEFF)));
}

/*draws the text for the custom menu*/
static void CustomMenuText(GAME_SETTINGS settings)
{
	/*buffer for sprintf*/
	char buffer[50];

	DrawStr(calc(29, 69), 0, "Use \x15\x16 to select, \x17\x18 to adjust", A_NORMAL);
	DrawStr(calc(42, 82), 6, "Press [ESC] to go back", A_NORMAL);
	sprintf(buffer, "Height: %u  ", settings.height);
	DrawStr(0, 13, buffer, A_REPLACE);
	sprintf(buffer, " Width: %u ", settings.width);
	DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 13, buffer, A_REPLACE);
	sprintf(buffer, "    Mines: %u", settings.mines);
	DrawStr(calc(159, 239) - DrawStrWidth(buffer, F_4x6), 13, buffer, A_REPLACE);
}

static void CustomMenuHighlight(GAME_SETTINGS settings)
{
	/*a buffer for sprintf*/
	char buffer[50];
	/*the left and right sides of the place to be highlighted*/
	short x0 = 0;
	short x1 = 0;

	switch(settings.column)
	{
		case 0:
		sprintf(buffer, "Height: %u", settings.height);
		x0 = 0;
		x1 = DrawStrWidth(buffer, F_4x6);
		break;

		case 1:
		sprintf(buffer, "Width: %u", settings.width);
		x0 = (calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2;
		x1 = (calc(160, 240) + DrawStrWidth(buffer, F_4x6)) / 2;
		break;

		case 2:
		sprintf(buffer, "Mines: %u", settings.mines);
		x0 = calc(159, 239) - DrawStrWidth(buffer, F_4x6);
		x1 = calc(159, 239);
	}

	FastFilledRect_Invert_R(LCD_MEM, x0, 12, x1, 18);
}

/*allows the user to adjust the key repeat delay*/
static void KeyDelayMenu(GAME_SETTINGS *settings)
{
	unsigned char column = 0;
	/*a varibale for use in rowread*/
	unsigned short key;

	ClrScr();

	KeyDelayMenuText(*settings);
	KeyDelayMenuTestStrip();
	KeyDelayMenuHighlight(column);

	do
	{
		key = _rowread(0xFFFE);

		/*up*/
		if(key & calc(1, 32))
		{
			settings->delay++;
		}

		/*down*/
		if(key & calc(4, 128))
		{
			settings->delay--;
		}

		/*up or down*/
		if(key & calc(5, 160))
		{
			/*redraw the text of the menu and wait for the current delay*/
			KeyDelayMenuText(*settings);
			settings->hw == 1 ? SleepMS_HW1(settings->delay) : SleepMS_HW2(settings->delay);
		}

		/*right or left*/
		if(key & calc(10, 80))
		{
			/*erase the highlight on the test strip*/
			KeyDelayMenuHighlight(column);
		}

		/*left*/
		if(key & calc(2, 16))
		{
			if(column == 0)
			{
				column = calc(25, 38);
			}
			else
			{
				column--;
			}
		}

		/*right*/
		if(key & calc(8, 64))
		{
			if(column == calc(25, 38))
			{
				column = 0;
			}
			else
			{
				column++;
			}
		}

		/*right or left*/
		if(key & calc(10, 80))
		{
			/*restore the highlight and wait for the specified delay*/
			KeyDelayMenuHighlight(column);
			settings->hw == 1 ? SleepMS_HW1(settings->delay) : SleepMS_HW2(settings->delay);
		}
		/*loop until [ESC] pressed*/
	} while(!(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)));
	/*wait until [ESC] released*/
	while(_rowread(calc(0xFFBF, 0xFEFF)));
}

/*draws the text for the key delay adjustment menu*/
static void KeyDelayMenuText(GAME_SETTINGS settings)
{
	/*a buffer for sprintf*/
	char buffer[50];

	DrawStr(calc(42, 82), 0, "Use \x17\x18 to adjust delay", A_NORMAL);
	DrawStr(calc(18, 58), 6, "Use \x15\x16 to test out the current delay", A_NORMAL);
	DrawStr(calc(42, 82), 12, "Press [ESC] to go back", A_NORMAL);
	sprintf(buffer, "  Delay: %ums  ", settings.delay);
	DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 24, buffer, A_REPLACE);
}

/*draws a strip to test out the key delay on*/
static void KeyDelayMenuTestStrip(void)
{
	FontSetSys(F_6x8);
	DrawStr(calc(2, 3), calc(50, 64), calc("12345678901234567890123456",
	"123456789012345678901234567890123456789"), A_NORMAL);
	FontSetSys(F_4x6);
}

/*draws the highlight on the test strip*/
static void KeyDelayMenuHighlight(unsigned char column)
{
	FastFilledRect_Invert_R(LCD_MEM, calc(2, 3) + 6 * column, calc(49, 63), calc(8, 9) + 6 * column,
	calc(57, 71));
}

/*sets up the width, height, and number of mines for the various game types*/
static void SetupGameParameters(unsigned char mines[], GAME_SETTINGS *settings, GAME_STATE *state)
{
	if(settings->scoring == SCORING_SCORED)
	{
		state->level++;

		switch(state->level)
		{
			case 0:
			settings->height = 6;
			settings->width = 10;
			settings->mines = 6;
			break;

			case 1:
			settings->height = 6;
			settings->width = 12;
			settings->mines = 7;
			break;

			case 2:
			settings->height = 6;
			settings->width = 14;
			settings->mines = 8;
			break;

			case 3:
			settings->height = 6;
			settings->width = 16;
			settings->mines = 9;
			break;

			case 4:
			settings->height = 8;
			settings->width = 16;
			settings->mines = 12;
			break;

			case 5:
			settings->height = 8;
			settings->width = 18;
			settings->mines = 14;
			break;

			case 6:
			settings->height = 8;
			settings->width = 20;
			settings->mines = 16;
			break;

			case 7:
			settings->height = 8;
			settings->width = 22;
			settings->mines = 17;
			break;

			case 8:
			settings->height = 8;
			settings->width = 24;
			settings->mines = 19;
			break;

			case 9:
			settings->height = 10;
			settings->width = 24;
			settings->mines = 24;
			break;

			case 10:
			settings->height = 10;
			settings->width = 26;
			settings->mines = 26;
			break;

			case 11:
			settings->height = calc(14, 10);
			settings->width = calc(20, 28);
			settings->mines = 28;
			break;

			case 12:
			settings->height = calc(15, 10);
			settings->width = calc(20, 30);
			settings->mines = 30;
			break;

			default:
			settings->height = calc(16, 10);
			settings->width = calc(21, 34);
			settings->mines = 24;
		}
	}
	else
	{
		Timer.time.ms = 0;
		Timer.time.sec = 0;
		Timer.time.min = 0;

		switch(settings->difficulty)
		{
			case DIFFICULTY_BEGINNER:
			settings->height = 9;
			settings->width = 9;
			settings->mines = 10;
			break;

			case DIFFICULTY_INTERMEDIATE:
			settings->height = 16;
			settings->width = 16;
			settings->mines = 40;
			break;

			case DIFFICULTY_ADVANCED:
			settings->height = 16;
			settings->width = calc(26, 30);
			settings->mines = calc(85, 99);
			break;

			case DIFFICULTY_PROGRESSIVE:
			settings->height = MAX_HEIGHT;
			settings->width = MAX_WIDTH;
			settings->mines = mines[settings->movement + settings->numbering] / 2;
			break;

			case DIFFICULTY_CUSTOM:
			break;
		}
	}
}

/*puts the mines and numbers into the minefield and keeps track of some things to tell when you have won*/
static void SetupMinefield(FLAGS minefield[][MAX_WIDTH], GAME_SETTINGS settings, GAME_STATE *state)
{
	short i;
	short j;
	short k;
	LOCATION temp;
	LOCATION temp2;

	/*make sure the minefield is clear of any old debris*/
	memset(minefield, 0, MAX_WIDTH * MAX_HEIGHT * sizeof(FLAGS));

	for(i = 0; i < (signed)settings.mines; i++)
	{
		do
		{
			temp = (LOCATION){random_long(settings.width), random_long(settings.height)};
			/*don't select the upper left corner unless you are playing a scored game, don't select the
			lower right corner when you are playing a game in walking mode, and don't put two mines in the
			same square*/
		} while((temp.x == 0 && temp.y == 0 && settings.scoring != SCORING_SCORED) || (temp.x ==
		settings.width - 1 && temp.y == settings.height - 1 && settings.movement == MOVEMENT_WALKING) ||
		minefield[temp.y][temp.x].mine == 1);

		minefield[temp.y][temp.x].mine = 1;
		minefield[temp.y][temp.x].adjacent = 0;

		/*put the numbers around the square, making sure to account for toroidal or not*/
		for(j = (settings.numbering == NUMBERING_PLANE && temp.x == 0 ? 0 : -1); j <= (settings.numbering
		== NUMBERING_PLANE && temp.x == settings.width - 1 ? 0 : 1); j++)
		{
			for(k = (settings.numbering == NUMBERING_PLANE && temp.y == 0 ? 0 : -1); k <=
			(settings.numbering == NUMBERING_PLANE && temp.y == settings.height - 1 ? 0 : 1); k++)
			{
				if(j != 0 || k != 0)
				{
					temp2 = temp;

					temp2.x += j;
					temp2.y += k;

					/*make sure the coordinate is still within the grid*/
					Wrap(&temp2, settings);

					if(!minefield[temp2.y][temp2.x].mine)
					{
						minefield[temp2.y][temp2.x].adjacent++;
					}
				}
			}
		}
	}

	/*count the number of non-zero squares for the winning condition in flying mode*/
	for(i = 0; i < settings.width; i++)
	{
		for(j = 0; j < settings.height; j++)
		{
			if(minefield[j][i].adjacent != 0)
			{
				state->nonZeroSquares++;
			}
		}
	}
}

/*makes sure that the coordinates given by location are within the game grid*/
static void Wrap(LOCATION *location, GAME_SETTINGS settings)
{
	if(location->x < 0)
	{
		location->x = settings.width - 1;
	}
	else if(location->x == settings.width)
	{
		location->x = 0;
	}

	if(location->y < 0)
	{
		location->y = settings.height - 1;
	}
	else if(location->y == settings.height)
	{
		location->y = 0;
	}
}

/*the function that starts the actual game*/
static GAME PlayGame(BOOL *savedGame, FLAGS minefield[][MAX_WIDTH], GAME_SETTINGS settings,
					 GAME_STATE *state, char backgroundSave[][18][7], QUEUE *q, unsigned char mines[])
{
	char inputRetVal;
	char checkRetVal = CHECK_MISS;

	/*make sure that grayscale is truned on*/
	if(!GrayOn())
	{
		return GAME_NO_GRAY;
	}

	GrayClearScreen();

	DrawMinefield(settings, minefield);

	/*have to automatically check the first square in walking mode*/
	if(settings.movement == MOVEMENT_WALKING)
	{
		CheckSquare(minefield, settings, state, q);
	}

	/*start the timer*/
	if(settings.scoring == SCORING_TIMED)
	{
		SetIntVec(AUTO_INT_5, Int5Timer);
	}

	DrawInfoBox(settings, *state, backgroundSave);

	do
	{
		inputRetVal = GetInput(settings, state, backgroundSave);

		EraseInfoBox(settings, *state, backgroundSave);

		switch(inputRetVal)
		{
			case INPUT_CHECK:
			checkRetVal = CheckSquare(minefield, settings, state, q);
			break;

			case INPUT_FLAG:
			DoFlag(minefield, settings, state);
			break;

			case INPUT_CHECKAROUND:
			checkRetVal = DoCheckAround(minefield, settings, state, q);
			break;

			case INPUT_SAVE:
			/*stop the timer*/
			SetIntVec(AUTO_INT_5, DUMMY_HANDLER);
			/*mark the data file as containing a saved game*/
			*savedGame = TRUE;
			GrayOff();
			return GAME_SAVE;
		}

		/*if there are entries in the autoclear queue that need to be taken care of*/
		if(q->front != q->back)
		{
			checkRetVal = DoAutoClear(minefield, settings, state, q);
		}

		if(inputRetVal != INPUT_ESC)
		{
			DrawInfoBox(settings, *state, backgroundSave);
		}

		if(CheckForWin(settings, *state))
		{
			if(settings.scoring != SCORING_SCORED)
			{
				/*stop the timer*/
				SetIntVec(AUTO_INT_5, DUMMY_HANDLER);
				if(settings.difficulty == DIFFICULTY_PROGRESSIVE)
				{
					mines[settings.movement + settings.numbering]++;
				}
				/*make sure all the squares with mines also have flags*/
				for(state->cursor.x = 0; state->cursor.x < settings.width; state->cursor.x++)
				{
					for(state->cursor.y = 0; state->cursor.y < settings.height; state->cursor.y++)
					{
						if(!minefield[state->cursor.y][state->cursor.x].flag)
						{
							DoFlag(minefield, settings, state);
						}
					}
				}

				/*draw the you win text*/
				FastFilledRect_Erase_R(GrayGetPlane(DARK_PLANE), calc(48, 88), calc(44, 58),
				calc(110, 150), calc(55, 69));
				FontSetSys(F_8x10);
				GrayDrawStr(calc(48,88), calc(45,59), "YOU WIN!", A_NORMAL);
				FontSetSys(F_4x6);

				/*wait for any key to be released, pressed, and released*/
				while(_rowread(0));
				while(!_rowread(0));
				while(_rowread(0));
			}
			else
			{
				/*set stuff up for the next scored level*/
				state->flags = 0;
				state->cursor = (LOCATION){0, 0};
				state->score += 10 * settings.mines;
				state->squaresCleared = 0;
			}

			GrayOff();
			return GAME_WIN;
		}
	} while(inputRetVal != INPUT_ESC && checkRetVal != CHECK_HIT);

	/*stop the timer*/
	SetIntVec(AUTO_INT_5, DUMMY_HANDLER);
	GrayOff();
	return GAME_LOSE;
}

/*draws the initial condition of the miefield*/
static void DrawMinefield(GAME_SETTINGS settings, FLAGS minefield[][MAX_WIDTH])
{
	short i;
	short j;

	DrawBorder(settings);

	for(i = 0; i < settings.width; i++)
	{
		for(j = 0; j < settings.height; j++)
		{
			if(minefield[j][i].checked)
			{
				DrawSprite(minefield[j][i].adjacent, settings, (LOCATION){i, j});
			}
			else
			{
				if(minefield[j][i].flag)
				{
					DrawSprite(SPRITE_FLAG, settings, (LOCATION){i, j});
				}
				else
				{
					DrawSprite(SPRITE_SQUARE, settings, (LOCATION){i, j});
				}
			}
		}
	}
}

/*draws the border around the minefield*/
static void DrawBorder(GAME_SETTINGS settings)
{
	GrayDrawRect((calc(157, 237) - 6 * settings.width) / 2, (calc(97, 125) - 6 * settings.height) / 2,
	(calc(161, 241) + 6 * settings.width) / 2, (calc(101, 129) + 6 * settings.height) / 2,
	COLOR_LIGHTGRAY, RECT_EMPTY);
}

/*takes care of most of the drawing of graphics*/
static void DrawSprite(SPRITE spriteNum, GAME_SETTINGS settings, LOCATION coords)
{
	short x;
	short y;
	static unsigned char sprite_0[5] = {};
	static unsigned char sprite_1[5] = {0x20, 0x60, 0x20, 0x20, 0x70};
	static unsigned char sprite_2[5] = {0x60, 0x10, 0x20, 0x40, 0x70};
	static unsigned char sprite_3[5] = {0x60, 0x10, 0x20, 0x10, 0x60};
	static unsigned char sprite_4[5] = {0x40, 0x50, 0x70, 0x10, 0x10};
	static unsigned char sprite_5[5] = {0x70, 0x40, 0x60, 0x10, 0x60};
	static unsigned char sprite_6[5] = {0x30, 0x40, 0x70, 0x50, 0x70};
	static unsigned char sprite_7[5] = {0x70, 0x10, 0x20, 0x40, 0x40};
	static unsigned char sprite_8[5] = {0x70, 0x50, 0x70, 0x50, 0x70};
	static unsigned char sprite_square[2][5] = {{0xF8, 0x88, 0x88, 0x88, 0xF8}, {0, 0x78, 0x78, 0x78,
	0x78}};
	static unsigned char sprite_mine[5] = {0x20, 0xA8, 0x70, 0xA8, 0x20};
	static unsigned char sprite_mine_hit[5] = {0xF8, 0xF8, 0xF8, 0xF8, 0xF8};
	static unsigned char sprite_flag[2][5] = {{0x20, 0x60, 0x20, 0x20, 0x20}, {0x60, 0xA0, 0x60, 0x20,
	0x20}};
	static unsigned char sprite_cursor[7] = {0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE};

	x = 6 * coords.x + (calc(161, 241) - 6 * settings.width) / 2;
	y = 6 * coords.y + (calc(101, 129) - 6 * settings.height) / 2;

	switch(spriteNum)
	{
		case SPRITE_0:
		GraySprite8_BLIT_R(x, y, 5, sprite_0, sprite_0, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_1:
		GraySprite8_BLIT_R(x, y, 5, sprite_1, sprite_1, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_2:
		GraySprite8_BLIT_R(x, y, 5, sprite_2, sprite_2, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_3:
		GraySprite8_BLIT_R(x, y, 5, sprite_3, sprite_3, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_4:
		GraySprite8_BLIT_R(x, y, 5, sprite_4, sprite_4, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_5:
		GraySprite8_BLIT_R(x, y, 5, sprite_5, sprite_5, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_6:
		GraySprite8_BLIT_R(x, y, 5, sprite_6, sprite_6, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_7:
		GraySprite8_BLIT_R(x, y, 5, sprite_7, sprite_7, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_8:
		GraySprite8_BLIT_R(x, y, 5, sprite_8, sprite_8, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_SQUARE:
		GraySprite8_BLIT_R(x, y, 5, sprite_square[0], sprite_square[1], 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_MINE:
		GraySprite8_BLIT_R(x, y, 5, sprite_mine, sprite_mine, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_MINE_HIT:
		GraySprite8_BLIT_R(x, y, 5, sprite_mine_hit, sprite_mine, 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_FLAG:
		GraySprite8_BLIT_R(x, y, 5, sprite_flag[0], sprite_flag[1], 0x07, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
		break;

		case SPRITE_CURSOR:
		GraySprite8_XOR_R(x - 1, y - 1, 7, sprite_cursor, sprite_cursor, GrayGetPlane(LIGHT_PLANE),
		GrayGetPlane(DARK_PLANE));
	}
}

/*draws the info box*/
static void DrawInfoBox(GAME_SETTINGS settings, GAME_STATE state, char backgroundSave[][18][7])
{
	/*a buffer for sprintf*/
	char buffer[30];
	SCR_RECT location;

	/*figure out the rectangle that the info box should go in*/
	location = LocateInfoBox(settings, state);

	Timer.coords = (LOCATION){location.xy.x0, location.xy.y0 + 7};

	/*save the background where the info box will go*/
	SpriteX8Get_R(location.xy.x0, location.xy.y0, min(18, calc(100, 128) - location.xy.y0),
	GrayGetPlane(LIGHT_PLANE), (char *)backgroundSave[0], 7);
	SpriteX8Get_R(location.xy.x0, location.xy.y0, min(18, calc(100, 128) - location.xy.y0),
	GrayGetPlane(DARK_PLANE), (char *)backgroundSave[1], 7);

	/*erase the dark plane to achieve a semi-transparent effect*/
	FastFilledRect_Erase_R(GrayGetPlane(DARK_PLANE), location.xy.x0, location.xy.y0, location.xy.x1,
	location.xy.y1);

	sprintf(buffer, "Mines Left: %u", settings.mines - state.flags);
	GrayDrawStr(location.xy.x0, location.xy.y0 + 1, buffer, A_NORMAL);
	if(settings.scoring == SCORING_TIMED)
	{
		/*save the background where the timer will go*/
		Sprite32Get_R(Timer.coords.x, Timer.coords.y, 5, GrayGetPlane(LIGHT_PLANE),
		((TIMER *)&Timer)->backgroundSave);
		PrintTime();
	}
	if(settings.scoring == SCORING_SCORED)
	{
		sprintf(buffer, "Level: %u", state.level + 1);
		GrayDrawStr(location.xy.x0, location.xy.y0 + 7, buffer, A_NORMAL);
		sprintf(buffer, "Score: %lu", state.score);
		GrayDrawStr(location.xy.x0, location.xy.y0 + 13, buffer, A_NORMAL);
	}

	/*turn the display of the timer back on*/
	Timer.print = TRUE;
}

static SCR_RECT LocateInfoBox(GAME_SETTINGS settings, GAME_STATE state)
{
	/*a buffer for sprintf*/
	char buffer[50];
	short width;
	short height;

	height = 6 + 6 * settings.scoring;

	sprintf(buffer, "Mines Left: %u", settings.mines - state.flags);
	width = DrawStrWidth(buffer, F_4x6);

	switch(settings.scoring)
	{
		case SCORING_NONE:
		break;

		case SCORING_TIMED:
		sprintf(buffer, "%02u:%02u", Timer.time.min, Timer.time.sec);
		width = max(width, DrawStrWidth(buffer, F_4x6));
		break;

		case SCORING_SCORED:
		sprintf(buffer, "%u", state.level + 1);
		width = max(width, DrawStrWidth(buffer, F_4x6));
		sprintf(buffer, "Score: %lu", state.score);
		width = max(width, DrawStrWidth(buffer, F_4x6));
		break;
	}

	if(state.cursor.x < settings.width / 2)
	{
		if(state.cursor.y < settings.height / 2)
		{
			/*bottom right*/
			return (SCR_RECT){{calc(159, 239) - width, calc(99, 127) - height, calc(159, 239), calc(99,
			127)}};
		}
		else
		{
			/*top right*/
			return (SCR_RECT){{calc(159, 239) - width, 0, calc(159, 239), height}};
		}
	}
	else
	{
		if(state.cursor.y < settings.height / 2)
		{
			/*bottom lef*/
			return (SCR_RECT){{0, calc(99, 127) - height, width, calc(99, 127)}};
		}
		else
		{
			/*top left*/
			return (SCR_RECT){{0, 0, width, height}};
		}
	}
}

static void EraseInfoBox(GAME_SETTINGS settings, GAME_STATE state, char backgroundSave[][18][7])
{
	/*a variable to hold the mask for the blit routine*/
	char mask[7] = {};
	/*how many pixels the max width hangs over the right edge*/
	short clipWidth;
	short i;
	short j;
	SCR_RECT location;

	/*disable the timer printing routine to give time to switch the coordinates of where it should print*/
	Timer.print = FALSE;

	location = LocateInfoBox(settings, state);

	clipWidth = max(54 - calc(160, 240) + location.xy.x0, 0);

	if(clipWidth > 0)
	{
		mask[6 - clipWidth / 8] = 0xFF >> (8 - clipWidth % 8);
		for(; clipWidth >= 8; clipWidth -= 8)
		{
			mask[7 - clipWidth / 8] = 0xFF;
		}
	}

	/*mask off any pixels in the background save that hang over the edge*/
	for(i = 0; i < 18; i++)
	{
		for(j = 0; j < 7; j++)
		{
			backgroundSave[0][i][j] &= ~mask[j];
			backgroundSave[1][i][j] &= ~mask[j];
		}
	}

	/*draw the saved background with clipping on the bottom and right edges only*/
	GraySpriteX8_BLIT(location.xy.x0, location.xy.y0, min(18, calc(100, 128) - location.xy.y0),
	(char *)backgroundSave[0], (char *)backgroundSave[1], mask, 7, GrayGetPlane(LIGHT_PLANE),
	GrayGetPlane(DARK_PLANE));
}

/*move the cursor around and decide what operation to do to the current square*/
static INPUT GetInput(GAME_SETTINGS settings, GAME_STATE *state, char backgroundSave[][18][7])
{
	/*a variable for rowread*/
	unsigned short key;

	/*draw the cursor*/
	DrawSprite(SPRITE_CURSOR, settings, state->cursor);

	do
	{
		key = _rowread(0xFFFE);

		/*any arrow*/
		if(key & calc(15, 240))
		{
			/*erase the cursor*/
			DrawSprite(SPRITE_CURSOR, settings, state->cursor);
		}

		/*up*/
		if(key & calc(1, 32))
		{
			/*if you cross the halfway point the info box needs to be moved*/
			if(state->cursor.y == settings.height / 2)
			{
				EraseInfoBox(settings, *state, backgroundSave);
				state->cursor.y--;
				DrawInfoBox(settings, *state, backgroundSave);
				if(settings.movement == MOVEMENT_WALKING)
				{
					/*wait until up released*/
					while(_rowread(0xFFFE));
					return INPUT_CHECK;
				}
			}
			else if(state->cursor.y == 0)
			{
				/*if you wrap over the edge the info box needs to be moved*/
				if(settings.numbering == NUMBERING_TORUS)
				{
					EraseInfoBox(settings, *state, backgroundSave);
					state->cursor.y = settings.height - 1;
					DrawInfoBox(settings, *state, backgroundSave);
				}
			}
			else
			{
				state->cursor.y--;
				if(settings.movement == MOVEMENT_WALKING)
				{
					/*wait until up released*/
					while(_rowread(0xFFFE));
					return INPUT_CHECK;
				}
			}
		}

		/*left*/
		if(key & calc(2, 16))
		{
			/*if you cross the halfway point the info box needs to be moved*/
			if(state->cursor.x == settings.width / 2)
			{
				EraseInfoBox(settings, *state, backgroundSave);
				state->cursor.x--;
				DrawInfoBox(settings, *state, backgroundSave);
				if(settings.movement == MOVEMENT_WALKING)
				{
					/*wait until left released*/
					while(_rowread(0xFFFE));
					return INPUT_CHECK;
				}
			}
			else if(state->cursor.x == 0)
			{
				/*if you wrap over the edge the info box needs to be moved*/
				if(settings.numbering == NUMBERING_TORUS)
				{
					EraseInfoBox(settings, *state, backgroundSave);
					state->cursor.x = settings.width - 1;
					DrawInfoBox(settings, *state, backgroundSave);
				}
			}
			else
			{
				state->cursor.x--;
				if(settings.movement == MOVEMENT_WALKING)
				{
					/*wait until left released*/
					while(_rowread(0xFFFE));
					return INPUT_CHECK;
				}
			}
		}

		/*down*/
		if(key & calc(4, 128))
		{
			/*if you cross the halfway point the info box needs to be moved*/
			if(state->cursor.y == settings.height / 2 - 1)
			{
				EraseInfoBox(settings, *state, backgroundSave);
				state->cursor.y++;
				DrawInfoBox(settings, *state, backgroundSave);
				if(settings.movement == MOVEMENT_WALKING)
				{
					/*wait until down released*/
					while(_rowread(0xFFFE));
					return INPUT_CHECK;
				}
			}
			else if(state->cursor.y == settings.height - 1)
			{
				/*if you wrap over the edge the info box needs to be moved*/
				if(settings.numbering == NUMBERING_TORUS)
				{
					EraseInfoBox(settings, *state, backgroundSave);
					state->cursor.y = 0;
					DrawInfoBox(settings, *state, backgroundSave);
				}
			}
			else
			{
				state->cursor.y++;
				if(settings.movement == MOVEMENT_WALKING)
				{
					/*wait until down released*/
					while(_rowread(0xFFFE));
					return INPUT_CHECK;
				}
			}
		}

		/*right*/
		if(key & calc(8, 64))
		{
			/*if you cross the halfway point the info box needs to be moved*/
			if(state->cursor.x == settings.width / 2 - 1)
			{
				EraseInfoBox(settings, *state, backgroundSave);
				state->cursor.x++;
				DrawInfoBox(settings, *state, backgroundSave);
				if(settings.movement == MOVEMENT_WALKING)
				{
					/*wait until right released*/
					while(_rowread(0xFFFE));
					return INPUT_CHECK;
				}
			}
			else if(state->cursor.x == settings.width - 1)
			{
				/*if you wrap over the edge the info box needs to be moved*/
				if(settings.numbering == NUMBERING_TORUS)
				{
					EraseInfoBox(settings, *state, backgroundSave);
					state->cursor.x = 0;
					DrawInfoBox(settings, *state, backgroundSave);
				}
			}
			else
			{
				state->cursor.x++;
				if(settings.movement == MOVEMENT_WALKING)
				{
					/*wait until right released*/
					while(_rowread(0xFFFE));
					return INPUT_CHECK;
				}
			}
		}

		/*any arrow*/
		if(key & calc(15, 240))
		{
			/*redraw the cursor and wait for the specified delay*/
			DrawSprite(SPRITE_CURSOR, settings, state->cursor);
			settings.hw == 1 ? SleepMS_HW1(settings.delay) : SleepMS_HW2(settings.delay);
		}

		/*2nd or F1*/
		if(calc(key, _rowread(0xFFBF)) & 16)
		{
			/*wait for 2nd or F1 to be released*/
			while(_rowread(calc(0xFFFE, 0xFFBF)));
			/*erase the cursor*/
			DrawSprite(SPRITE_CURSOR, settings, state->cursor);

			return INPUT_CHECK;
		}

		/*diamond or F2*/
		if(calc(key, _rowread(0xFFEF)) & calc(64, 16))
		{
			/*wait for diamond or F2 to be released*/
			while(_rowread(calc(0xFFFE, 0xFFEF)));
			/*erase the cursor*/
			DrawSprite(SPRITE_CURSOR, settings, state->cursor);

			return INPUT_FLAG;
		}

		/*shift or F3*/
		if(calc(key, _rowread(0xFFFB)) & calc(32, 16))
		{
			/*wait for shift or F3 to be released*/
			while(_rowread(calc(0xFFFE, 0xFFFB)));
			/*erase the cursor*/
			DrawSprite(SPRITE_CURSOR, settings, state->cursor);

			return INPUT_CHECKAROUND;
		}

		/*clear*/
		if(_rowread(calc(0xFFFD, 0xFF7F)) & calc(64, 32))
		{
			/*wait for clear to be released*/
			while(_rowread(calc(0xFFFD, 0xFF7F)));
			/*erase the cursor*/
			DrawSprite(SPRITE_CURSOR, settings, state->cursor);

			return INPUT_SAVE;
		}
		/*loop until [ESC] pressed*/
	} while(!(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)));
	/*wait until [ESC] released*/
	while(_rowread(calc(0xFFBF, 0xFEFF)));

	return INPUT_ESC;
}

/*check to see if you hit a mine or not and handle associated graphics*/
static CHECK CheckSquare(FLAGS minefield[][MAX_WIDTH], GAME_SETTINGS settings, GAME_STATE *state,
						 QUEUE *q)
{
	/*if you havent checked or flagged the current square*/
	if(minefield[state->cursor.y][state->cursor.x].checked != 1 &&
	minefield[state->cursor.y][state->cursor.x].flag != 1)
	{
		/*if you hit a mine*/
		if(minefield[state->cursor.y][state->cursor.x].mine == 1)
		{
			/*draw all the mines*/
			DrawMines(minefield, settings, *state);

			/*draw you lose with transparent effect*/
			FastFilledRect_Erase_R(GrayGetPlane(DARK_PLANE), calc(48, 88), calc(44, 58), calc(112, 152),
			calc(55, 69));
			FontSetSys(F_8x10);
			GrayDrawStr(calc(48,88), calc(45,59), "YOU LOSE", A_NORMAL);
			FontSetSys(F_4x6);
			while(_rowread(0));
			while(!_rowread(0));
			while(_rowread(0));
			return CHECK_HIT;
		}
		else
		{
			/*draw the graphics for the square*/
			DrawSprite(minefield[state->cursor.y][state->cursor.x].adjacent, settings, state->cursor);

			/*if you hti an empty square in non-walking mode*/
			if(minefield[state->cursor.y][state->cursor.x].adjacent == 0 && settings.movement !=
			MOVEMENT_WALKING)
			{
				/*update some info for telling when the game is won*/
				state->emptySquaresCleared++;
				/*add the square to the autoclear queue*/
				q->data[q->back] = state->cursor;
				q->back = (q->back + 1) % QUEUE_SIZE;
			}

			minefield[state->cursor.y][state->cursor.x].checked = 1;

			/*update some more info for telling when the game is won*/
			state->squaresCleared++;

			state->score++;
		}
	}

	return CHECK_MISS;
}

/*draw the mines when you hit one*/
static void DrawMines(FLAGS minefield[][MAX_WIDTH], GAME_SETTINGS settings, GAME_STATE state)
{
	short x;
	short y;

	for(x = 0; x < settings.width; x++)
	{
		for(y = 0; y < settings.height; y++)
		{
			if(minefield[y][x].mine)
			{
				if(x == state.cursor.x && y == state.cursor.y)
				{
					/*highlight the mine you hit*/
					DrawSprite(SPRITE_MINE_HIT, settings, (LOCATION){x, y});
				}
				else
				{
					DrawSprite(SPRITE_MINE, settings, (LOCATION){x, y});
				}
			}
		}
	}
}

/*indicates that you think there is a mine in the current square*/
static void DoFlag(FLAGS minefield[][MAX_WIDTH], GAME_SETTINGS settings, GAME_STATE *state)
{
	/*if you haven't checked the square*/
	if(!minefield[state->cursor.y][state->cursor.x].checked)
	{
		/*if the square is flagged*/
		if(minefield[state->cursor.y][state->cursor.x].flag)
		{
			/*turn the flag off*/
			minefield[state->cursor.y][state->cursor.x].flag = 0;
			state->flags--;
			DrawSprite(SPRITE_SQUARE, settings, state->cursor);
		}
		else
		{
			/*turn the flag on*/
			minefield[state->cursor.y][state->cursor.x].flag = 1;
			state->flags++;
			DrawSprite(SPRITE_FLAG, settings, state->cursor);
		}
	}
}

/*checks around a square if the number of flags adjacent to the square equals the generated number of
mines next to the square*/
static CHECK DoCheckAround(FLAGS minefield[][MAX_WIDTH], GAME_SETTINGS settings, GAME_STATE *state,
						   QUEUE *q)
{
	/*the number of flags adjacent to the square*/
	char flags = 0;
	short i;
	short j;
	/*save a copy of the current location so that it can be modified without messing things up*/
	LOCATION temp = state->cursor;

	/*if the square is checked and has some mines adjacent to it*/
	if(minefield[state->cursor.y][state->cursor.x].checked &&
	minefield[state->cursor.y][state->cursor.x].adjacent != 0)
	{
		/*add up the number of adjacent flags*/
		for(i = (settings.numbering == NUMBERING_PLANE && temp.x == 0 ? 0 : -1); i <= (settings.numbering
		== NUMBERING_PLANE && temp.x == settings.width - 1 ? 0 : 1); i++)
		{
			for(j = (settings.numbering == NUMBERING_PLANE && temp.y == 0 ? 0 : -1); j <=
			(settings.numbering == NUMBERING_PLANE && temp.y == settings.height - 1 ? 0 : 1); j++)
			{
				state->cursor = temp;

				state->cursor.x += i;
				state->cursor.y += j;

				Wrap(&state->cursor, settings);

				flags += minefield[state->cursor.y][state->cursor.x].flag;
			}
		}

		/*restore the cursor location*/
		state->cursor = temp;

		/*if the number of flags and the number of mines match*/
		if(flags == minefield[state->cursor.y][state->cursor.x].adjacent)
		{
			/*check all squares adjacent to the current square*/
			for(i = (settings.numbering == NUMBERING_PLANE && temp.x == 0 ? 0 : -1); i <=
			(settings.numbering == NUMBERING_PLANE && temp.x == settings.width - 1 ? 0 : 1); i++)
			{
				for(j = (settings.numbering == NUMBERING_PLANE && temp.y == 0 ? 0 : -1); j <=
				(settings.numbering == NUMBERING_PLANE && temp.y == settings.height - 1 ? 0 : 1); j++)
				{
					state->cursor = temp;

					state->cursor.x += i;
					state->cursor.y += j;

					Wrap(&state->cursor, settings);

					/*oops! the flags weren't in the right place*/
					if(CheckSquare(minefield, settings, state, q) == CHECK_HIT)
					{
						return CHECK_HIT;
					}
				}
			}
		}
		else
		{
			/*flash the squares that would have been checked*/
			for(i = (settings.numbering == NUMBERING_PLANE && temp.x == 0 ? 0 : -1); i <=
			(settings.numbering == NUMBERING_PLANE && temp.x == settings.width - 1 ? 0 : 1); i++)
			{
				for(j = (settings.numbering == NUMBERING_PLANE && temp.y == 0 ? 0 : -1); j <=
				(settings.numbering == NUMBERING_PLANE && temp.y == settings.height - 1 ? 0 : 1); j++)
				{
					state->cursor = temp;

					state->cursor.x += i;
					state->cursor.y += j;

					Wrap(&state->cursor, settings);

					if(minefield[state->cursor.y][state->cursor.x].checked == 0)
					{
						DrawSprite(SPRITE_CURSOR, settings, state->cursor);
					}
				}
			}

			settings.hw == 1 ? SleepMS_HW1(settings.delay) : SleepMS_HW2(settings.delay);

			for(i = (settings.numbering == NUMBERING_PLANE && temp.x == 0 ? 0 : -1); i <=
			(settings.numbering == NUMBERING_PLANE && temp.x == settings.width - 1 ? 0 : 1); i++)
			{
				for(j = (settings.numbering == NUMBERING_PLANE && temp.y == 0 ? 0 : -1); j <=
				(settings.numbering == NUMBERING_PLANE && temp.y == settings.height - 1 ? 0 : 1); j++)
				{
					state->cursor = temp;

					state->cursor.x += i;
					state->cursor.y += j;

					Wrap(&state->cursor, settings);

					if(minefield[state->cursor.y][state->cursor.x].checked == 0)
					{
						DrawSprite(SPRITE_CURSOR, settings, state->cursor);
					}
				}
			}
		}

		/*restore the cursor again*/
		state->cursor = temp;
	}

	return CHECK_MISS;
}

/*if you hit a empty square, all squares around it are checked. if any of those are empty, the process
repreats*/
static CHECK DoAutoClear(FLAGS minefield[][MAX_WIDTH], GAME_SETTINGS settings, GAME_STATE *state,
						 QUEUE *q)
{
	char checkRetVal;
	short i;
	short j;
	LOCATION backup = state->cursor;
	LOCATION temp;

	do
	{
		/*take the first square in line*/
		temp = q->data[q->front];

		/*check all the squares next to it*/
		for(i = (settings.numbering == NUMBERING_PLANE && temp.x == 0 ? 0 : -1); i <= (settings.numbering
		== NUMBERING_PLANE && temp.x == settings.width - 1 ? 0 : 1); i++)
		{
			for(j = (settings.numbering == NUMBERING_PLANE && temp.y == 0 ? 0 : -1); j <=
			(settings.numbering == NUMBERING_PLANE && temp.y == settings.height - 1 ? 0 : 1); j++)
			{
				if(i != 0 || j != 0)
				{
					state->cursor = temp;

					state->cursor.x += i;
					state->cursor.y += j;

					Wrap(&state->cursor, settings);

					checkRetVal = CheckSquare(minefield, settings, state, q);

					/*this shouldn't happen*/
					if(checkRetVal == CHECK_HIT)
					{
						return CHECK_HIT;
					}
				}
			}
		}

		/*move the front of the line back one spot*/
		q->front = (q->front + 1) % QUEUE_SIZE;
		/*loop until the line is empty*/
	} while(q->front != q->back);

	/*restore the cursor*/
	state->cursor = backup;

	return CHECK_MISS;
}

/*checks to see if you have won the game*/
static BOOL CheckForWin(GAME_SETTINGS settings, GAME_STATE state)
{
	if(settings.scoring != SCORING_SCORED)
	{
		if(settings.movement == MOVEMENT_WALKING)
		{
			if(state.cursor.x == settings.width - 1 && state.cursor.y == settings.height - 1)
			{
				/*HP mode won*/
				return TRUE;
			}
		}
		else
		{
			if(state.squaresCleared - state.emptySquaresCleared == state.nonZeroSquares)
			{
				/*Windows style won*/
				return TRUE;
			}
		}
	}
	else
	{
		if(state.squaresCleared == (unsigned)(settings.width * settings.height - settings.mines) &&
		state.flags == settings.mines)
		{
			/*scored mode won*/
			return TRUE;
		}
	}

	/*didn't win yet*/
	return FALSE;
}

/*checks to see if you got a high score*/
static void CheckHighScores(GAME_SETTINGS settings, HIGH_SCORES *highScores, unsigned long score)
{
	if(settings.scoring == SCORING_SCORED)
	{
		if(score > highScores->scores[settings.numbering])
		{
			FontSetSys(F_8x10);
			GrayDrawStr(calc(36,76), calc(45,59), "HIGH SCORE!", A_REPLACE);
			FontSetSys(F_4x6);
			while(_rowread(0));
			while(!_rowread(0));
			while(_rowread(0));

			highScores->scores[settings.numbering] = score;
		}
	}
	else
	{
		if(memcmp((TIME *)(&Timer.time), &highScores->times[settings.movement +
		settings.numbering][settings.difficulty], sizeof(TIME)) < 0)
		{
			FontSetSys(F_8x10);
			GrayDrawStr(calc(40,80), calc(45,59), "BEST TIME!", A_REPLACE);
			FontSetSys(F_4x6);
			while(_rowread(0));
			while(!_rowread(0));
			while(_rowread(0));

			memcpy(&highScores->times[settings.movement + settings.numbering][settings.difficulty],
			(TIME *)(&Timer.time), sizeof(TIME));
		}
	}
}

/*prints the time*/
static void PrintTime(void)
{
	char buffer[10];

	/*restores the light plane and erases the dark plane for that nifty semi-transparent effect*/
	GraySprite32_RPLC_R(((TIMER *)&Timer)->coords.x, ((TIMER *)&Timer)->coords.y, 5,
	((TIMER *)&Timer)->backgroundSave, (unsigned long[]){0, 0, 0, 0, 0}, GrayGetPlane(LIGHT_PLANE),
	GrayGetPlane(DARK_PLANE));

	sprintf(buffer, "%02u:%02u", ((TIMER *)&Timer)->time.min, ((TIMER *)&Timer)->time.sec);
	GrayDrawStr(((TIMER *)&Timer)->coords.x, ((TIMER *)&Timer)->coords.y, buffer, A_NORMAL);
}